/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifdef __GNUG__
#pragma implementation
#endif

#include "yoo_root.hxx"
#include "dictionary.hxx"

#include <string.h>

char command[10000] ;


ADM_Server::ADM_Server(char* n) {

  cerr << "\t-> Create Server " << n << endl ;

  this->name = n ;
  this->log = YOO_SYSTEM_DIR ;
  YoodaDictionary->dictServ.insert(this) ;

  cerr << "\t<- Create Server " << n << endl ;
}


ADM_Server::~ADM_Server() {

  cerr << "\t-> Delete Server " << this->name << endl ;

  YoodaDictionary->dictServ.remove(this) ;
  Iterator<ADM_Vol> iter(this->l_vol) ;
  for (; iter.more(); iter.next()) {
    delete (ADM_Vol*)iter.get() ;
  }

  cerr << "\t<- Delete Server " << this->name << endl ;
}


void ADM_Server::movelog(char* name) {

  cerr << "\t-> Move log for Server " << this->name << endl ;

  this->log = name ;

  cerr << "\t<- Move log for Server " << this->name << endl ;
}


void ADM_Server::print(int lev, char* head) {

  cerr << "\n" << head << "*** Display Server ***" << endl ;
  cerr << head << "\tName   = " << this->name << endl ;

  if (lev > 0) {
    int card = this->l_vol.count() ;
    if (card > 0) {
      cerr << head << "\tVolumes :" << endl ; 
    }
    char subHead[100] ;
    subHead[0] = 0 ;
    strcpy(subHead, head) ;
    strcat(subHead,"\t") ;
    for (int i = 0; i < card; i++) {
      this->l_vol[i]->print(lev-1, subHead) ;
    }
  }
}


ADM_Vol::ADM_Vol(char* vol, char* f, Ref<ADM_Server> s) {

  cerr << "\t-> Create Volume " << vol << endl ;

  this->name = vol ;
  if (f == NULL) {
    this->file.assign(YOO_VOL_DIR,"/",vol) ;
  } else {
    this->file = f ;
  }
  if (s == NULL)
    this->server = YoodaDictionary->defServ ;
  else 
    this->server = s ;

  this->server->l_vol += this ;
  this->volId = YoodaDictionary->lastVol + 1 ;
  YoodaDictionary->lastVol += 1 ;
  YoodaDictionary->dictVol.insert(this) ;

  cerr << "\t<- Create Volume " << vol << endl ;
} 


ADM_Vol::~ADM_Vol() {

  cerr << "\t-> Delete Volume " << this->name << endl ;

  YoodaDictionary->dictVol.remove(this) ;

  Iterator<ADM_Clu> iter(this->l_clu) ;
  for (; iter.more(); iter.next()) {
    delete (ADM_Clu*)iter.get() ;
  }
  
  int pos = this->server->l_vol.member(this) ;
  this->server->l_vol.remove(pos) ;
  
  strcat(command,"\nrm ") ;
  strcat(command, this->file) ;

  cerr << "\t<- Delete Volume " << this->name << endl ;
}


int ADM_Vol::movefile(char* f) {

  cerr << "\t-> Move volume file" << this->name << endl ;

  char buff[512] ;
  sprintf(buff,"\ncp %s %s\nrm %s",(char*)this->file,f,(char*)this->file) ;
  strcat(command, buff) ;

  this->file = f ;

  cerr << "\t<- Move Volume " << this->name << endl ;
  return 1 ;
}


int ADM_Vol::moveserver(Ref<ADM_Server> s) {

  cerr << "\t-> Move volume " << this->name << " in server" << s->name << endl;

  int pos = this->server->l_vol.member(this) ;
  this->server->l_vol.remove(pos) ;
  
  this->server = s ;
  this->server->l_vol += this ;

  cerr << "\t<- Move Volume " << this->name << " in server" << s->name << endl;
  return 1 ;
}


void ADM_Vol::print(int lev, char* head) {

  cerr << head << "\n*** Display " ;
  if (this->volId < ADM_DefaultVol) {
    cerr << head << " System Volume ***" << endl ;
  } else {
    cerr << head << " User Volume ***" << endl ;
  }
  cerr << head << "\tName = " << this->name << endl ;
  cerr << head << "\tfile = " << this->file << endl ;
  if (lev > 0) {
    int card = this->l_clu.count() ;
    if (card > 0) {
      cerr << head << "\tClusters :" << endl ; 
    }
    char subHead[100] ;
    subHead[0] = 0 ;
    strcpy(subHead, head) ;
    strcat(subHead,"\t") ;
    for (int i = 0; i < card; i++) {
      this->l_clu[i]->print(lev-1, subHead) ;
    }
  }
}  


ADM_Clu::ADM_Clu(char* n, ADM_Vol* vol) {

  cerr << "\t-> Create Cluster in " << vol->name << endl ;

  this->name = n ;

  this->clu = new Cluster(vol->volId) ;
  this->myVol = vol ;
  this->myVol->l_clu += this ;

  YoodaDictionary->dictClu.insert(this) ;

  cerr << "\t<- Create Cluster in " << vol->name << endl ;
}


ADM_Clu::~ADM_Clu() {

  cerr << "\t-> Delete Cluster " << this->name << endl ;

  YoodaDictionary->dictClu.remove(this) ;
  ListX<ADM_Class> l_temp = this->l_class ;
  for (int i = 0; i < l_temp.count(); i++) {
    l_temp[i]->move(YoodaDictionary->defClu) ;
  }

  int pos = this->myVol->l_clu.member(this) ;
  this->myVol->l_clu.remove(pos) ;
  
  cerr << "\t<- Delete Cluster " << this->name << endl ;
}


void ADM_Clu::print(int lev, char* head) {

  cerr << "\n" << head << "*** Display Cluster ***" << endl ;
  cerr << head << "\tName   = " << this->name << endl ;
  cerr << head << "\tVolume = " << this->myVol->name << endl ;

  if (lev > 0) {
    int card = this->l_class.count() ;
    if (card > 0) {
      cerr << head << "\tClasses :" << endl ; 
    }
    char subHead[100] ;
    subHead[0] = 0 ;
    strcpy(subHead, head) ;
    strcat(subHead,"\t") ;
    for (int i = 0; i < card; i++) {
      this->l_class[i]->print(subHead) ;
    }
  }
}
  

ADM_Class::ADM_Class(char* clname, char* sup, ADM_Clu* clu, int id) {

  cerr << "\t-> Create Class " << clname << endl ;  

  this->name = clname ;
  this->super = sup ;
  this->myClu = clu ;
  this->myClu->l_class += this ;

  if (id == -1)
    this->clId = YoodaDictionary->lastClass ;
  else 
    this->clId = id;

  if ( YoodaDictionary->dictClass.insertUniq(this) ) {
    YoodaDictionary->lastClass += 1 ;
  }
}


ADM_Class::~ADM_Class() {
  
  cerr << "\t-> Delete Class " << this->name << endl ;  
  int pos = this->myClu->l_class.member(this) ;
  this->myClu->l_class.remove(pos) ;
  YoodaDictionary->dictClass.remove(this) ;
}


int ADM_Class::supernumber()
{
  int num = 1 ;
  ADM_Class* curr = this ;
  while (curr->superclass(1) != NULL) {
    curr = curr->superclass(1) ;
    num++ ;
  }
  return num ;
}


ADM_Class* ADM_Class::superclass(int num)
{
  if (this->super == NULL) return NULL ;
  ADM_Class* curr = YoodaDictionary->dictClass.getFirst(this->super) ;
  if (curr == NULL) {
    cerr << this->super << " : Unknown class" << endl ;
  }
  for (int i = 1; i < num; i++) {    
    curr = curr->superclass(1) ;
  }
  return curr ;
}


void ADM_Class::yoodaGenRef()
{
  char file[512] ;
  
  // Create ref_file
  sprintf(file, "%s/ref_%s.hxx", YOO_SCH_DIR, (char*)this->name) ;
  FILE* f_ref = fopen(file, "w") ;
  fprintf(f_ref, "#ifndef __%s_HXX\n", (char*)this->name) ;
  fprintf(f_ref, "#define __%s_HXX\n", (char*)this->name) ;

  fprintf(f_ref, "\n#ifndef __yoo_root_HXX\n") ;
  fprintf(f_ref, "#include \"yoo_root.hxx\"\n") ;
  fprintf(f_ref, "#endif\n") ;

  fprintf(f_ref, "\n\nclass %s ;\n", (char*)this->name) ;
  fprintf(f_ref, "class Ref<%s> : public yoo_ObjectX {\n", (char*)this->name) ;
  fprintf(f_ref, "  public :\n") ;
  fprintf(f_ref, "  Ref() {}\n") ;
  fprintf(f_ref, "  Ref(OID oid) : yoo_ObjectX(oid) {} ;\n") ;
  fprintf(f_ref, "  Ref(%s* e) : yoo_ObjectX(e) {} ;\n", (char*)this->name) ;
  fprintf(f_ref, "  %s* operator->() const {\n", (char*)this->name) ;
  fprintf(f_ref, "    PageHdlMem* pd ;\n") ;
  fprintf(f_ref, "    return (%s*)((pd=(PageHdlMem*)YOO_VirtualMemory->_inVMPagePID[_oid])!=NULL ? pd->_page->object(_oid.objNo) : YOO_VirtualServer->findPage(this->_oid)->object(_oid.objNo)) ;\n", (char*)this->name) ;
  fprintf(f_ref, "  }\n") ;
  fprintf(f_ref, "  %s& operator*() const {\n", (char*)this->name) ;
  fprintf(f_ref, "    return *this->operator->() ;\n") ;
  fprintf(f_ref, "  }\n") ;
  fprintf(f_ref, "  operator %s*() const {\n", (char*)this->name) ;
  fprintf(f_ref, "    return this->_oid==NULL_OID?NULL:this->operator->();\n");
  fprintf(f_ref, "  }\n") ;
  fprintf(f_ref, "  class Ref<%s>& operator=(%s* obj) {\n", (char*)this->name, (char*)this->name) ;
  fprintf(f_ref, "    this->assign(obj); return *this;\n") ;
  fprintf(f_ref, "  }\n") ;

  for (int i = 1; i < this->supernumber(); i++) {
    fprintf(f_ref, "  operator %s*() const { return (%s*)getObj() ; }\n", (char*)this->superclass(i)->name, (char*)this->superclass(i)->name) ;
    fprintf(f_ref, "  operator Ref<%s>() const { return Ref<%s>(this->_oid) ; }\n", (char*)this->superclass(i)->name, (char*)this->superclass(i)->name) ;
  }
  fprintf(f_ref, "} ;\n") ;
  fprintf(f_ref, "\nyoo_declare(ListX<%s>)\n",(char*)this->name) ;
  fprintf(f_ref, "\n\n#endif\n\n") ;
  fclose(f_ref) ;
}


void ADM_Class::yoodaGenStub(FILE* f_out)
{
  this->yoodaGenRef() ;
  fprintf(f_out, "\n\n  public :\n") ;
  fprintf(f_out, "  static int yoo_ClId;\n") ;
  fprintf(f_out, "  static Ref<Cluster> yoo_Cluster;\n") ;
  fprintf(f_out, "  void* operator new(size_t s,void* addr)\n") ;
  fprintf(f_out, "  {\n") ;
  fprintf(f_out, "    return addr;\n") ;
  fprintf(f_out, "  }\n") ;
  fprintf(f_out, "  void* operator new(size_t s,Cluster* c=%s::yoo_Cluster)\n",
	  (char*)this->name) ;
  fprintf(f_out, "  {\n") ;
  fprintf(f_out, "    return ObjectX::operator new(s,c,%s::yoo_ClId);\n",
	  (char*)this->name) ;
  fprintf(f_out, "  }\n") ;
  fprintf(f_out, "  void* operator new(size_t s,ObjectX* o)\n") ;
  fprintf(f_out, "  {\n") ;
  fprintf(f_out, "    return ObjectX::operator new(s,o,%s::yoo_ClId);\n",
	  (char*)this->name) ;
  fprintf(f_out, "  }\n") ;
  fprintf(f_out, "  static void yoo_refresh(void* addr) \n") ;
  fprintf(f_out, "  {\n") ;
  fprintf(f_out, "    YOO_LOAD_TIME = TRUE ;\n") ;
  fprintf(f_out, "    new(addr) %s ;\n", (char*)this->name) ;
  fprintf(f_out, "    YOO_LOAD_TIME = FALSE ;\n") ;
  fprintf(f_out, "  }\n") ;
}


void ADM_Class::move(ADM_Clu* cluster) {

  cerr << "\t-> Move Class " << this->name << endl ;

  int pos = this->myClu->l_class.member(this) ;
  this->myClu->l_class.remove(pos) ;
  this->myClu = cluster ;
  this->myClu->l_class += this ;

  cerr << "\t<- Move Class " << this->name << endl ;
}

void ADM_Class::print(char* head) {

  cerr << "\n" << head << "*** Display Class ***" << endl ;
  cerr << head << "\tName    = " << this->name << endl ;
  cerr << head << "\tClassId = " << this->clId << endl ;
  cerr << head << "\tCluster = " << this->myClu->name << endl ;
  cerr << head << "\tVolume  = " << this->myClu->myVol->name << endl ;
}


ADM_NameObj::ADM_NameObj(StringX n, yoo_ObjectX o) {
  this->name = n ;
  this->obj = o ;
  this->clId = o->getClass() ;
  ADM_NameObj* old = YoodaDictionary->dictName.getFirst(n) ;
  if (old != NULL) {
    call_throw(YOO_ERR_ADM, "Name already used") ;
  }
  YoodaDictionary->dictName.insert(this) ;
}


ADM_NameObj::~ADM_NameObj() {
  
  YoodaDictionary->dictName.remove(this) ;
} 


void ADM_NameObj::print(char* head) {
  
  cerr << "\n" << head << "*** Display Name ***" << endl ;
  cerr << head << "\tName    = " << this->name << endl ;
  Iterator<ADM_Class> iter(YoodaDictionary->dictClass);
  while (iter.more()) {
    if (iter.get()->clId == this->clId) {
      cerr << head << "\tClass = " << iter.get()->name << endl ;
      break ;
    }
    iter.next() ;
  }
}


ADM_Admin::ADM_Admin() {

  this->lastVol = 0 ;
  this->defVol = NULL ;
  this->defClu = NULL ;
  this->defServ = NULL ;
  this->lastClass = ADM_FirstClId ; // first unused class Id
}


int yoo_setname(yoo_ObjectX obj, const StringX& str) 
{
  ADM_NameObj* no = YoodaDictionary->dictName.getFirst(str) ;
  if (no != NULL) {
    YoodaDictionary->dictName.remove(no) ;
  }
  new ADM_NameObj(str, obj) ;
  return TRUE ;
}


int yoo_getname(yoo_ObjectX& obj, const StringX& str) 
{
  ADM_NameObj* no = YoodaDictionary->dictName.getFirst(str) ;
  if (no == NULL) {
    obj = NULL ;
    return FALSE ;
  } else {
    obj = no->obj ;
    return TRUE ;
  }
}


int yoo_delname(const StringX& str) 
{
  YoodaDictionary->dictName.removeAll(str) ;
  return TRUE ;
}

void genVol() {

  FILE* f_vol = fopen(YOO_VOL_FILE, "w") ;
  if (f_vol == NULL) {
    perror ("Error : Cannot open volume file") ;
    exit(1) ;
  }

  Iterator<ADM_Server> iter(YoodaDictionary->dictServ) ;
  for (; iter.more(); iter.next()) {
    fprintf(f_vol, "server %s {\n", (char*)iter.get()->name) ;
    fprintf(f_vol, "\tlog %s ;\n", (char*)iter.get()->log) ;
    Iterator<ADM_Vol> itervol(iter.get()->l_vol) ;
    for (; itervol.more(); itervol.next()) {
      fprintf(f_vol, "\tvolume %s {\n", (char*)itervol.get()->name) ;
      fprintf(f_vol, "\t\tident %d ;\n",(int)itervol.get()->volId ) ;
      fprintf(f_vol, "\t\tfile %s ;\n",(char*)itervol.get()->file ) ;
      fprintf(f_vol, "\t}\n") ;
    }
    fprintf(f_vol, "}\n") ;
  }

  fclose(f_vol) ;
}

